from __future__ import annotations

from typing import Optional

from msgspec import Raw, Struct, field


class RealData:
    """
    Real data wrapper class.

    """

    def __init__(self, values: bytes, type: str, name: str, item: str):
        """
        Initialize RealData instance.

        'values' can be used as bytes itself.
            values_str = values.decode('utf-8')
            values_dic = orjson.loads(values)
            values_dic = msgspec.json.decode(values)

        Args:
            values (bytes): utf-8 encoded bytes
            type (str): API type (ex. OB, 0D, ...)
            name (str): API name (ex. 주식체결, 주식호가잔량, ...)
            item (str): stock code
        """
        self.values = values  # utf-8 encoded bytes
        self.type = type  # API type (ex. OB, 0D, ...)
        self.name = name  # API name (ex. 주식체결, 주식호가잔량, ...)
        self.item = item  # stock code

    def __repr__(self):
        return (
            f"RealData(values={self.values}, type={self.type}, name={self.name}, item={self.item})"
        )


# To decode raw json with msgspec
class RealType(Struct):
    class Data(Struct):
        values: Raw
        type: str
        name: str
        item: str

    trnm: str
    data: Optional[list[RealType.Data]] = None


# To decode raw json with msgspec
class TickValuesType(Struct):
    """
    To decode Real.data.values of type 'OB'
    """

    # Real.Data.values
    v20: str = field(name="20")  # 체결시간
    v10: str = field(name="10")  # 현재가
    v15: str = field(name="15")  # +-거래량


# To decode raw json with msgspec
class HogaValuesType(Struct):
    """
    To decode Real.data.values of type '0D'
    """

    # Real.Data.Values
    v21: str = field(name="21")  # 호가시간

    v41: str = field(name="41")  # 매도호가 1
    v61: str = field(name="61")  # 매도잔량 1
    v51: str = field(name="51")  # 매수호가 1
    v71: str = field(name="71")  # 매수잔량 1

    v42: str = field(name="42")  # 매도호가 2
    v62: str = field(name="62")  # 매도잔량 2
    v52: str = field(name="52")  # 매수호가 2
    v72: str = field(name="72")  # 매수잔량 2

    v43: str = field(name="43")  # 매도호가 3
    v63: str = field(name="63")  # 매도잔량 3
    v53: str = field(name="53")  # 매수호가 3
    v73: str = field(name="73")  # 매수잔량 3

    v44: str = field(name="44")  # 매도호가 4
    v64: str = field(name="64")  # 매도잔량 4
    v54: str = field(name="54")  # 매수호가 4
    v74: str = field(name="74")  # 매수잔량 4

    v45: str = field(name="45")  # 매도호가 5
    v65: str = field(name="65")  # 매도잔량 5
    v55: str = field(name="55")  # 매수호가 5
    v75: str = field(name="75")  # 매수잔량 5

    v46: str = field(name="46")  # 매도호가 6
    v66: str = field(name="66")  # 매도잔량 6
    v56: str = field(name="56")  # 매수호가 6
    v76: str = field(name="76")  # 매수잔량 6

    v47: str = field(name="47")  # 매도호가 7
    v67: str = field(name="67")  # 매도잔량 7
    v57: str = field(name="57")  # 매수호가 7
    v77: str = field(name="77")  # 매수잔량 7

    v48: str = field(name="48")  # 매도호가 8
    v68: str = field(name="68")  # 매도잔량 8
    v58: str = field(name="58")  # 매수호가 8
    v78: str = field(name="78")  # 매수잔량 8

    v49: str = field(name="49")  # 매도호가 9
    v69: str = field(name="69")  # 매도잔량 9
    v59: str = field(name="59")  # 매수호가 9
    v79: str = field(name="79")  # 매수잔량 9

    v50: str = field(name="50")  # 매도호가 10
    v70: str = field(name="70")  # 매도잔량 10
    v60: str = field(name="60")  # 매수호가 10
    v80: str = field(name="80")  # 매수잔량 10


class Types:
    TICK = {
        "20": int,  # 체결시간
        "10": int,  # 현재가
        "15": int,  # +-거래량
    }

    HOGA = {
        "21": int,  # 호가시간
        "41": int,  # 매도호가 1
        "61": int,  # 매도잔량 1
        "51": int,  # 매수호가 1
        "71": int,  # 매수잔량 1
        "42": int,  # 매도호가 2
        "62": int,  # 매도잔량 2
        "52": int,  # 매수호가 2
        "72": int,  # 매수잔량 2
        "43": int,  # 매도호가 3
        "63": int,  # 매도잔량 3
        "53": int,  # 매수호가 3
        "73": int,  # 매수잔량 3
        "44": int,  # 매도호가 4
        "64": int,  # 매도잔량 4
        "54": int,  # 매수호가 4
        "74": int,  # 매수잔량 4
        "45": int,  # 매도호가 5
        "65": int,  # 매도잔량 5
        "55": int,  # 매수호가 5
        "75": int,  # 매수잔량 5
        "46": int,  # 매도호가 6
        "66": int,  # 매도잔량 6
        "56": int,  # 매수호가 6
        "76": int,  # 매수잔량 6
        "47": int,  # 매도호가 7
        "67": int,  # 매도잔량 7
        "57": int,  # 매수호가 7
        "77": int,  # 매수잔량 7
        "48": int,  # 매도호가 8
        "68": int,  # 매도잔량 8
        "58": int,  # 매수호가 8
        "78": int,  # 매수잔량 8
        "49": int,  # 매도호가 9
        "69": int,  # 매도잔량 9
        "59": int,  # 매수호가 9
        "79": int,  # 매수잔량 9
        "50": int,  # 매도호가 10
        "70": int,  # 매도잔량 10
        "60": int,  # 매수호가 10
        "80": int,  # 매수잔량 10
    }



import asyncio
import warnings
from itertools import chain
from os.path import isfile
from typing import Callable

import aiohttp
from aiohttp import ClientSession

from kiwoom.config.http import (
    HTTP_CONNECT_TIMEOUT,
    HTTP_READ_TIMEOUT,
    HTTP_TCP_CONNECTORS,
    HTTP_TOTAL_TIMEOUT,
    State,
)
from kiwoom.http.debug import debugger, dumps
from kiwoom.http.response import Response
from kiwoom.http.utils import RateLimiter

__all__ = ["Client"]


class Client:
    def __init__(self, host: str, appkey: str, secretkey: str):
        """
        Initialize Client instance.

        Args:
            host (str): domain
            appkey (str): file path or raw appkey
            secretkey (str): file path or raw secretkey
        """
        self.host: str = host
        self.debugging: bool = False

        self._auth: str = ""
        self._appkey: str = appkey
        self._secretkey: str = secretkey

        self._state_http = State.CLOSED
        self._ready_event = asyncio.Event()
        self._limiter: RateLimiter = RateLimiter()
        self._session: ClientSession = None

    async def connect(self, appkey: str, secretkey: str) -> None:
        """
        Connect to Kiwoom REST API server and receive token.

        Args:
            appkey (str): file path or raw appkey
            secretkey (str): file path or raw secretkey
        """
        if isfile(appkey):
            with open(appkey, "r") as f:
                self._appkey = f.read().strip()
        if isfile(secretkey):
            with open(secretkey, "r") as f:
                self._secretkey = f.read().strip()

        # Already connected
        if self._session and not self._session.closed:
            return

        # Establish HTTP session
        self._ready_event.clear()
        self._session = ClientSession(
            timeout=aiohttp.ClientTimeout(
                total=HTTP_TOTAL_TIMEOUT,
                sock_connect=HTTP_CONNECT_TIMEOUT,
                sock_read=HTTP_READ_TIMEOUT,
            ),
            connector=aiohttp.TCPConnector(limit=HTTP_TCP_CONNECTORS, enable_cleanup_closed=True),
        )

        # Request token
        endpoint = "/oauth2/token"
        api_id = ""
        headers = self.headers(api_id)
        data = {
            "grant_type": "client_credentials",
            "appkey": self._appkey,
            "secretkey": self._secretkey,
        }
        async with self._session.post(self.host + endpoint, headers=headers, json=data) as res:
            res.raise_for_status()
            body = await res.json()
            resp = Response(res.url, res.status, res.headers, body)

        # Set token
        if "token" not in body:
            msg = dumps(self, endpoint, api_id, headers, data, resp)
            raise RuntimeError(f"Failed to get token: {msg}")
        token = body["token"]
        self._auth = f"Bearer {token}"
        self._session.headers.update(
            {
                "Content-Type": "application/json;charset=UTF-8",
                "authorization": self._auth,
            }
        )
        self._state_http = State.CONNECTED
        self._ready_event.set()

    async def close(self) -> None:
        """
        Close HTTP session.
        """
        self._ready_event.clear()
        if self._session:
            await asyncio.shield(self._session.close())

        self._auth = ""
        self._session = None

    def token(self) -> str:
        """
        Returns token if available, otherwise empty string.

        Raises:
            ValueError: Invalid token.

        Returns:
            str: token
        """
        if not self._auth:
            return ""
        if "Bearer " in self._auth:
            return self._auth[len("Bearer ") :]
        raise ValueError(f"Invalid token: {self._auth}")

    def headers(
        self, api_id: str, cont_yn: str = "N", next_key: str = "", headers: dict | None = None
    ) -> dict[str, str]:
        """
        Generate headers for the request.

        Args:
            api_id (str): api_id in Kiwoom API
            cont_yn (str, optional): cont_yn in Kiwoom API
            next_key (str, optional): next_key in Kiwoom API
            headers (dict | None, optional): headers to be updated with

        Returns:
            dict[str, str]: headers
        """
        base = {
            # 'Content-Type': 'application/json;charset=UTF-8',
            # 'authorization': self._auth,
            "cont-yn": cont_yn,
            "next-key": next_key,
            "api-id": api_id,
        }
        if headers is not None:
            headers.update(base)
            return headers
        return base

    async def ready(self):
        """
        Wait until request limit is lifted and connection is established.

        Raises:
            RuntimeError: Connection timeout.
        """
        try:
            await asyncio.wait_for(self._ready_event.wait(), HTTP_TOTAL_TIMEOUT)
        except asyncio.TimeoutError as err:
            msg = f"Connection timeout: waited for {HTTP_TOTAL_TIMEOUT} seconds."
            raise RuntimeError(msg) from err
        await self._limiter.acquire()

    @debugger
    async def post(
        self, endpoint: str, api_id: str, headers: dict | None = None, data: dict | None = None
    ) -> aiohttp.ClientResponse:
        """
        Post request to the server, but using client.request function is recommended.
        Request limit and connection status are checked globally and automatically.

        Args:
            endpoint (str): endpoint to Kiwoom REST API server
            api_id (str): api id
            headers (dict | None, optional): headers of the request.
            data (dict | None, optional): data to be sent in json format

        Returns:
            aiohttp.ClientResponse: async response from the server,
                but this will be converted to kiwoom.http.response.Response by debugger.
        """

        # Warn not connected
        if not self._state_http == State.CONNECTED:
            warnings.warn("Not connected, wait for timeout...", RuntimeWarning, stacklevel=1)

        # Wait connection and request limits
        await self.ready()

        # Post Request
        if headers is None:
            headers = self.headers(api_id)
        return await self._session.post(self.host + endpoint, headers=headers, json=data)

    async def request(
        self, endpoint: str, api_id: str, headers: dict | None = None, data: dict | None = None
    ) -> Response:
        """
        Requests to the server and returns response with error handling.

        Args:
            endpoint (str): endpoint of the server
            api_id (str): api id
            headers (dict | None, optional): headers of the request. Defaults to None.
            data (dict | None, optional): data of the request. Defaults to None.

        Raises:
            RuntimeError: RuntimeError when return_code is not in [0, 3, 20]

        Returns:
            Response: response wrapped by kiwoom.http.response.Response
        """

        res: Response = await self.post(endpoint, api_id, headers=headers, data=data)
        body = res.json()
        if "return_code" in body:
            match body["return_code"]:
                case 0 | 20:
                    # 0: Success
                    # 20 : No Data
                    return res
                case 3:
                    # 3 : Token Expired
                    print("Token expired, trying to refresh token...")
                    await self.connect(self._appkey, self._secretkey)
                    return await self.request(endpoint, api_id, headers=headers, data=data)

        # Request Failure
        return_code = body["return_code"]
        err = f"\nRequest failed with {return_code=}, not in {{'0', '3', '20'}}."
        if not self.debugging:
            msg = dumps(self, endpoint, api_id, headers, data, res)
            raise RuntimeError(msg + err)
        raise RuntimeError(err)

    async def request_until(
        self,
        should_continue: Callable,
        endpoint: str,
        api_id: str,
        headers: dict | None = None,
        data: dict | None = None,
    ) -> dict:
        """
        Request until 'cont-yn' in response header is 'Y',
        and should_continue(body) evaluates to True.

        Args:
            should_continue (Callable):
                callable that takes body(dict) and
                returns boolean value to request again or not
            endpoint (str):
                endpoint of the server
            api_id (str):
                api id
            headers (dict | None, optional):
                headers of the request. Defaults to None.
            data (dict | None, optional):
                data of the request. Defaults to None.

        Returns:
            dict: response body
        """

        # Initial request
        res = await self.request(endpoint, api_id, headers=headers, data=data)
        body = res.json()

        # If condition to chain is not met
        if callable(should_continue) and not should_continue(body):
            return body

        # Extract list data only
        bodies = dict()
        for key in body.keys():
            if isinstance(body[key], list):
                bodies[key] = [body[key]]
                continue
            bodies[key] = body[key]

        # Rercursive call
        while res.headers.get("cont-yn") == "Y" and should_continue(body):
            next_key = res.headers.get("next-key")
            headers = self.headers(api_id, cont_yn="Y", next_key=next_key, headers=headers)

            # Continue request
            res = await self.request(endpoint, api_id, headers=headers, data=data)
            body = res.json()

            # Append list data
            for key in body.keys():
                if isinstance(body[key], list):
                    bodies[key].append(body[key])

        # Flatten list data as if it was one list
        for key in bodies:
            if isinstance(bodies[key], list):
                bodies[key] = list(chain.from_iterable(bodies[key]))
        return bodies



from os import getcwd, makedirs
from os.path import exists, isdir, join
from pathlib import Path

from pandas import DataFrame

from kiwoom.config import ENCODING
from kiwoom.config.candle import (
    COLUMN_MAPPER_CANDLE,
    PERIOD_TO_BODY_KEY,
    PERIOD_TO_COLUMN,
    handle_time,
    valid,
)


# 캔들차트 관련 처리함수
def process(
    data: dict,
    code: str,
    period: str,
    ctype: str,
    start: str,
    end: str,
) -> DataFrame:
    columns: list[str] = PERIOD_TO_COLUMN[period]
    if not valid(data, period, ctype):
        # Returns empty dataframe
        df = DataFrame(columns=columns)
        return df

    key: str = PERIOD_TO_BODY_KEY[ctype][period]
    df = DataFrame(data[key])
    df = df[::-1]

    mapper = COLUMN_MAPPER_CANDLE[period]
    df.rename(columns=mapper, inplace=True)
    df = df[columns]

    time = columns[0]
    df = handle_time(df, code, period)
    df.set_index(time, drop=True, inplace=True)
    if not df.index.is_monotonic_increasing:
        df = df.sort_index(kind="stable")
    df = df.astype(int).abs()
    df = df.loc[start:end]
    return df


async def to_csv(file: str, path: str, df: DataFrame, encoding: str = ENCODING) -> None:
    # Validate
    if df.empty:
        return
    if not path:
        path = getcwd()
    if not isdir(path):
        raise ValueError(f"Path not valid: '{path}'")
    if not exists(path):
        makedirs(path)

    # Save
    if not file.endswith(".csv"):
        file = f"{file}.csv"
    file = join(path, file)
    if exists(file):
        Path.unlink(file)
    df.to_csv(file, encoding=encoding)



# 일반적인 처리함수
def stock_list(data: dict, ats: bool = True) -> list[str]:
    codes = list()
    if not ats:
        for dic in data['list']:
            codes.append(dic['code'])
        codes = sorted(codes)
        return codes

    # 대체거래소 통합시세 코드
    for dic in data['list']:
        if dic['nxtEnable'] == 'Y':
            codes.append(dic['code'] + '_AL')
            continue
        codes.append(dic['code'])
    codes = sorted(codes)
    return codes



import asyncio
import contextlib
from typing import Self

from pandas import DataFrame

from kiwoom import proc
from kiwoom.api import API
from kiwoom.config.http import EXCEPTIONS_TO_SUPPRESS


class Bot:
    """
    Kiwoom REST API를 이용해 전략을 실행하는 최상위 클래스입니다.

    사용자가 API 세부 동작을 알지 못해도 전략 수행에 집중할 수 있도록 합니다.
    """

    def __init__(self, host: str, appkey: str, secretkey: str, api: API | None = None):
        """
        Bot 클래스 인스턴스를 초기화합니다.

        Args:
            host (str): 실서버 / 모의서버 도메인
            appkey (str): 파일경로 / 앱키
            secretkey (str): 파일경로 / 시크릿키
            api (API, optional): API를 별도로 구현했다면 인스턴스 전달가능
        """
        self.api = api if api else API(host, appkey, secretkey)

    async def __aenter__(self) -> Self:
        """
        async with 구문에서 Bot 인스턴스를 반환합니다.

        Returns:
            Bot: self
        """
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        """
        async with 구문을 종료할 때 연결을 해제하고 리소스를 정리합니다.

        Args:
            exc_type (_type_): exception type
            exc_value (_type_): exception value
            traceback (_type_): traceback
        """
        with contextlib.suppress(EXCEPTIONS_TO_SUPPRESS):
            await asyncio.shield(self.close())

    def debug(self, debugging: bool = True) -> None:
        """
        디버깅 모드 활성화 / 비활성화.
        디버깅 모드에서는 Http 요청과 응답이 출력됩니다.

        Args:
            debugging (bool): 디버깅 모드 활성화 여부
        """
        self.api.debugging = debugging

    def token(self) -> str:
        """
        연결이 되었다면, 키움 REST API 토큰을 반환합니다.

        Returns:
            str: token
        """
        return self.api.token()

    async def connect(self):
        """
        키움 REST API HTTP 서버 및 Websocket 서버에 접속합니다.
        """
        await self.api.connect()
        await asyncio.sleep(1)

    async def close(self):
        """
        키움 REST API HTTP 서버 및 Websocket 서버 연결을 해제합니다.
        """
        await asyncio.shield(self.api.close())

    async def stock_list(self, market: str, ats: bool = True) -> list[str]:
        """
        주어진 market code에 해당하는 주식 종목코드 목록을 반환합니다.

        Args:
            market (str): {
                'KOSPI': '0', 'KOSDAQ': '10', 'ELW': '3',
                '뮤추얼펀드': '4', '신주인수권': '5', '리츠': '6',
                'ETF': '8', '하이일드펀드': '9', 'K-OTC': '30',
                'KONEX': '50', 'ETN': '60', 'NXT': 'NXT'}
            ats (bool, optional): 대체거래소 반영한 통합코드 여부 (ex. '005930_AL')

        Returns:
            list[str]: 종목코드 리스트
        """

        # Add NXT market
        if market == "NXT":
            kospi = await self.stock_list("0")
            kosdaq = await self.stock_list("10")
            codes = [c for c in kospi + kosdaq if "AL" in c]
            return sorted(codes)

        data = await self.api.stock_list(market)
        codes = proc.stock_list(data, ats)
        return codes

    async def candle(
        self,
        code: str,
        period: str,
        ctype: str,
        start: str = None,
        end: str = None,
    ) -> DataFrame:
        """
        주어진 코드, 기간, 종목/업종 유형에 해당하는 캔들차트 데이터를 반환합니다.

        Args:
            code (str): 종목코드 / 업종코드
            period (str): 캔들 기간유형 {"tick", "min", "day"}
            ctype (str): 종목 / 업종유형 {"stock", "sector"}
            start (str, optional): 시작일자 in YYYYMMDD format
            end (str, optional): 종료일자 in YYYYMMDD format

        Returns:
            DataFrame: Pandas 캔들차트 데이터프레임
        """
        data = await self.api.candle(code, period, ctype, start, end)
        df = proc.candle.process(data, code, period, ctype, start, end)
        return df

    async def trade(self, start: str, end: str = "") -> DataFrame:
        """
        주어진 시작일자와 종료일자에 해당하는 체결내역을
        키움증권 '0343' 계좌 체결내역 화면과 동일한 구성으로 반환합니다.
        데이터 조회 제한으로 최근 2개월 데이터만 조회할 수 있습니다.

        체결내역 데이터는 [알파노트](http://alphanote.io)를 통해
        간편하게 진입/청산 시각화 및 성과 지표들을 확인할 수 있습니다.

        Args:
            start (str): 시작일자 in YYYYMMDD format
            end (str, optional): 종료일자 in YYYYMMDD format

        Returns:
            DataFrame: 키움증권 '0343' 화면 'Excel 내보내기' 형식
        """
        data = await self.api.trade(start, end)
        df = proc.trade.process(data)
        return df

    async def run(self):
        """
        전략 로직을 구현하고 실행합니다.
        """
        pass



from os import getcwd, makedirs
from os.path import exists, isdir, join
from pathlib import Path

from pandas import DataFrame

from kiwoom.config import ENCODING
from kiwoom.config.trade import (
    COLUMN_MAPPER_TRADE,
    COLUMN_TRADE,
)


# 체결내역 관련 처리함수
def process(data: list[dict]) -> DataFrame:
    if not data:
        return DataFrame(columns=COLUMN_TRADE)

    df = DataFrame(data)
    df.rename(columns=COLUMN_MAPPER_TRADE, inplace=True)

    ints = [
        "주문번호",
        "원주문번호",
        "주문수량",
        "주문단가",
        "확인수량",
        "스톱가",
        "체결번호",
        "체결수량",
        "체결평균단가",
    ]
    for col in ints:
        df[col] = df[col].astype(int)

    df["주식채권"] = df["주식채권"].map({"1": "주식", "2": "채권"})
    df["원주문번호"] = df["원주문번호"].apply(lambda x: "" if x == 0 else str(x))
    df["종목번호"] = df["종목번호"].str[-6:]
    df["체결시간"] = df["체결시간"].str.lstrip("0")

    df = df[COLUMN_TRADE]
    return df


async def to_csv(file: str, path: str, df: DataFrame, encoding: str = ENCODING):
    # Validate
    if df.empty:
        print("DataFrame is empty, skip writing to csv.")
        return
    if not path:
        path = getcwd()
    if not isdir(path):
        raise ValueError(f"Path not valid: '{path}'")
    if not exists(path):
        makedirs(path)

    # Save
    if not file.endswith(".csv"):
        file = f"{file}.csv"
    file = join(path, file)
    if exists(file):
        Path.unlink(file)

    # 키움증권 0343 화면
    col1 = [  # 1st row
        "주식채권",
        "주문번호",
        "원주문번호",
        "종목번호",
        "매매구분",
        "주문유형구분",
        "주문수량",
        "주문단가",
        "확인수량",
        "체결번호",
        "스톱가",
    ]
    col2 = [  # 2nd row
        "주문일자",
        "종목명",
        "접수구분",
        "신용거래구분",
        "체결수량",
        "체결평균단가",
        "정정/취소",
        "통신",
        "예약/반대",
        "체결시간",
        "거래소",
    ]
    df = df.astype(str)
    with open(file, "w", encoding=encoding) as f:
        lines = []
        row1 = df[col1]
        row2 = df[col2]
        for (_, r1), (_, r2) in zip(row1.iterrows(), row2.iterrows(), strict=True):
            lines.append(",".join(r1) + "\n")
            lines.append(",".join(r2) + "\n")

        f.write(",".join(col1) + "\n")  # Header col 1
        f.write(",".join(col2) + "\n")  # Header col 2
        f.writelines(lines)



import functools
import json
from typing import Callable

from aiohttp import ClientResponse
from aiohttp.web import HTTPException

from kiwoom.http.response import Response

__all__ = ["dumps", "debugger"]


def dumps(api, endpoint: str, api_id, headers: dict, data: dict, res: Response) -> str:
    """
    Dump request and response to string for debugging.

    Args:
        api (Client): Client instance
        endpoint (str): endpoint
        api_id (str): api id
        headers (dict): headers
        data (dict): data
        res (Response): wrapped response

    Returns:
        str: _description_
    """
    # Request
    headers = json.dumps(
        headers if headers is not None else api.headers(api_id), indent=4, ensure_ascii=False
    )
    req = "\n== Request ==\n"
    req += f"URL : {api.host + endpoint}\n"
    req += f"Headers : {headers}\n"
    req += f"Data : {json.dumps(data, indent=4, ensure_ascii=False)}\n"

    # Response
    headers = json.dumps(
        {key: res.headers.get(key) for key in ["next-key", "cont-yn", "api-id"]},
        indent=4,
        ensure_ascii=False,
    )
    resp = "== Response ==\n"
    resp += f"Code : {res.status}\n"
    resp += f"Headers : {headers}\n"
    resp += f"Response : {json.dumps(res.json(), indent=4, ensure_ascii=False)}\n"
    return req + resp


def debugger(fn) -> Callable:
    """
    Debugger decorator for Client.post method.
    Even though debugging is disabled, it will print if error occurs.

    Args:
        fn (function): function to be decorated

    Raises:
        err: propagate HTTPException from original function

    Returns:
        Response: wrapped response
    """

    @functools.wraps(fn)
    async def wrapper(api, endpoint: str, api_id: str, headers: dict, data: dict) -> Response:
        res: ClientResponse = await fn(api, endpoint, api_id, headers, data)
        async with res:
            # Async to sync Response
            resp = Response(
                url=res.url, status=res.status, headers=res.headers, body=await res.json()
            )

            # Debugging
            if api.debugging:
                print(dumps(api, endpoint, api_id, headers, data, resp))

            try:
                res.raise_for_status()
            except HTTPException as err:
                # Always debug when error occurs
                if not api.debugging:
                    print(dumps(api, endpoint, api_id, headers, data, resp))
                raise err
        return resp

    return wrapper



REQUEST_LIMIT_DAYS = 60 + 5  # 2 months + extra days

COLUMN_TRADE: list[str] = [
    # 1st row
    '주식채권', '주문번호', '원주문번호', '종목번호', '매매구분',
    '주문유형구분', '주문수량', '주문단가', '확인수량', '체결번호', '스톱가',

    # 2nd row
    '주문일자', '종목명', '접수구분', '신용거래구분', '체결수량', '체결평균단가',
    '정정/취소', '통신', '예약/반대', '체결시간', '거래소'
]

COLUMN_MAPPER_TRADE: dict[str, str] = {
    # 1st row
    'stk_bond_tp': '주식채권',
    'ord_no': '주문번호',
    'orig_ord_no': '원주문번호',
    'stk_cd': '종목번호',
    'trde_tp': '매매구분',
    'io_tp_nm': '주문유형구분',
    'ord_qty': '주문수량',
    'ord_uv': '주문단가',
    'cnfm_qty': '확인수량',
    'cntr_no': '체결번호',
    'cond_uv': '스톱가',

    # 2nd row
    'ord_dt': '주문일자',
    'stk_nm': '종목명',
    'acpt_tp': '접수구분',
    'crd_deal_tp': '신용거래구분',
    'cntr_qty': '체결수량',
    'cntr_uv': '체결평균단가',
    'mdfy_cncl_tp': '정정/취소',
    'comm_ord_tp': '통신',
    'rsrv_oppo': '예약/반대',
    'cntr_tm': '체결시간',
    'dmst_stex_tp': '거래소',
}
